@@ -124,28 +124,28 @@ static void test_comparison_operators()
       "rule test { condition: 0.5 <= 1}", NULL);
 
   assert_true_rule(
-      "rule rest { condition: 1.0 <= 1}", NULL);
+      "rule test { condition: 1.0 <= 1}", NULL);
 
   assert_true_rule(
-      "rule rest { condition: \"abc\" == \"abc\"}", NULL);
+      "rule test { condition: \"abc\" == \"abc\"}", NULL);
 
   assert_true_rule(
-      "rule rest { condition: \"abc\" <= \"abc\"}", NULL);
+      "rule test { condition: \"abc\" <= \"abc\"}", NULL);
 
   assert_true_rule(
-      "rule rest { condition: \"abc\" >= \"abc\"}", NULL);
+      "rule test { condition: \"abc\" >= \"abc\"}", NULL);
 
   assert_true_rule(
-      "rule rest { condition: \"ab\" < \"abc\"}", NULL);
+      "rule test { condition: \"ab\" < \"abc\"}", NULL);
 
   assert_true_rule(
-      "rule rest { condition: \"abc\" > \"ab\"}", NULL);
+      "rule test { condition: \"abc\" > \"ab\"}", NULL);
 
   assert_true_rule(
-      "rule rest { condition: \"abc\" < \"abd\"}", NULL);
+      "rule test { condition: \"abc\" < \"abd\"}", NULL);
 
   assert_true_rule(
-      "rule rest { condition: \"abd\" > \"abc\"}", NULL);
+      "rule test { condition: \"abd\" > \"abc\"}", NULL);
 
   assert_false_rule(
       "rule test { condition: 1 != 1}", NULL);
@@ -355,6 +355,26 @@ static void test_syntax()
   assert_error(
       "rule test { strings: $a = \"a\" $a = \"a\" condition: all of them }",
       ERROR_DUPLICATED_STRING_IDENTIFIER);
+
+  assert_error(
+      "rule test { strings: $a = /a.c/ xor condition: $a }",
+      ERROR_SYNTAX_ERROR);
+
+  assert_error(
+      "rule test { strings: $a = /abc/ xor condition: $a }",
+      ERROR_SYNTAX_ERROR);
+
+  assert_error(
+      "rule test { strings: $a = {01 02 ?? 03 04} xor condition: $a }",
+      ERROR_SYNTAX_ERROR);
+
+  assert_error(
+      "rule test { strings: $a = {01 02 0? 03 04} xor condition: $a }",
+      ERROR_SYNTAX_ERROR);
+
+  assert_error(
+      "rule test { strings: $a = {01 02 03 04} xor condition: $a }",
+      ERROR_SYNTAX_ERROR);
 }
 
 
@@ -509,14 +529,16 @@ static void test_strings()
              all of them\n\
        }", "abcdef");
 
+  // xor by itself will match the plaintext version of the string too.
   assert_true_rule_file(
     "rule test {\n\
       strings:\n\
         $a = \"This program cannot\" xor\n\
       condition:\n\
-        #a == 255\n\
+        #a == 256\n\
     }", "tests/data/xor.out");
 
+  // Make sure the combination of xor and ascii behaves the same as just xor.
   assert_true_rule_file(
     "rule test {\n\
       strings:\n\
@@ -525,6 +547,25 @@ static void test_strings()
         #a == 256\n\
     }", "tests/data/xor.out");
 
+  assert_true_rule_file(
+    "rule test {\n\
+      strings:\n\
+        $a = \"This program cannot\" xor(1-0x10)\n\
+      condition:\n\
+        #a == 16\n\
+    }", "tests/data/xor.out");
+
+  // We should have no matches here because we are not generating the ascii
+  // string, just the wide one, and the test data contains no wide strings.
+  assert_true_rule_file(
+    "rule test {\n\
+      strings:\n\
+        $a = \"This program cannot\" xor wide\n\
+      condition:\n\
+        #a == 0\n\
+    }", "tests/data/xor.out");
+
+  // xor by itself is equivalent to xor(0-255).
   assert_true_rule_file(
     "rule test {\n\
       strings:\n\
@@ -533,13 +574,406 @@ static void test_strings()
         #a == 256\n\
     }", "tests/data/xorwide.out");
 
+  // This DOES NOT look for the plaintext wide version by itself.
   assert_true_rule_file(
+    "rule test {\n\
+      strings:\n\
+        $a = \"This program cannot\" xor(1-16) wide\n\
+      condition:\n\
+        #a == 16\n\
+    }", "tests/data/xorwide.out");
+
+  // Check the location of the match to make sure we match on the correct one.
+  assert_true_rule_file(
+    "rule test {\n\
+      strings:\n\
+        $a = \"This program cannot\" xor(1) wide\n\
+      condition:\n\
+        #a == 1 and @a == 0x2f\n\
+    }", "tests/data/xorwide.out");
+
+  assert_error(
+    "rule test {\n\
+      strings:\n\
+        $a = \"This program cannot\" xor(300)\n\
+      condition:\n\
+        $a\n\
+    }", ERROR_INVALID_MODIFIER);
+
+  assert_error(
+    "rule test {\n\
+      strings:\n\
+        $a = \"This program cannot\" xor(200-10)\n\
+      condition:\n\
+        $a\n\
+    }", ERROR_INVALID_MODIFIER);
+
+  assert_error(
+    "rule test {\n\
+      strings:\n\
+        $a = {00 11 22 33} xor\n\
+      condition:\n\
+        $a\n\
+    }", ERROR_SYNTAX_ERROR);
+
+  assert_error(
+    "rule test {\n\
+      strings:\n\
+        $a = /foo(bar|baz)/ xor\n\
+      condition:\n\
+        $a\n\
+    }", ERROR_SYNTAX_ERROR);
+
+  assert_error(
+    "rule test {\n\
+      strings:\n\
+        $a = \"ab\" xor xor\n\
+      condition:\n\
+        $a\n\
+    }", ERROR_DUPLICATED_MODIFIER);
+
+  // We should have no matches here because we are not generating the wide
+  // string, just the ascii one, and the test data contains no ascii strings.
+  assert_true_rule_file(
+    "rule test {\n\
+      strings:\n\
+        $a = \"This program cannot\" xor ascii\n\
+      condition:\n\
+        #a == 0\n\
+    }", "tests/data/xorwide.out");
+
+  // This should match 512 times because we are looking for the wide and ascii
+  // versions in plaintext and doing xor(0-255) (implicitly)
+  assert_true_rule_file(
+    "rule test {\n\
+      strings:\n\
+        $a = \"This program cannot\" xor wide ascii\n\
+      condition:\n\
+        #a == 512\n\
+    }", "tests/data/xorwideandascii.out");
+
+  assert_true_rule_file(
+    "rule test {\n\
+      strings:\n\
+        $a = \"This program cannot\" wide ascii\n\
+      condition:\n\
+        #a == 2\n\
+    }", "tests/data/xorwideandascii.out");
+
+  assert_error(
     "rule test {\n\
       strings:\n\
         $a = \"ab\" xor nocase\n\
       condition:\n\
-        #a == 1084\n\
-    }", "tests/data/xornocase.out");
+        true\n\
+    }", ERROR_INVALID_MODIFIER);
+
+  assert_true_rule(
+      "rule test { \
+        strings:\
+          $a = \"AXS\" private\
+      condition:\
+        all of them\
+      }",
+      "AXS");
+
+  assert_true_rule(
+      "rule test { \
+        strings:\
+          $a = { 45 52 53 } private\
+      condition:\
+        all of them\
+      }",
+      "ERS");
+
+  assert_true_rule(
+      "rule test { \
+        strings:\
+          $a = /AXS[0-9]{4}ERS[0-9]{4}/ private\
+      condition:\
+        all of them\
+      }",
+      "AXS1111ERS2222");
+
+  assert_error(
+    "rule test {\n\
+      strings:\n\
+        $a = \"ab\" base64 nocase\n\
+      condition:\n\
+        true\n\
+    }", ERROR_INVALID_MODIFIER);
+
+  assert_error(
+    "rule test {\n\
+      strings:\n\
+        $a = \"ab\" base64 xor\n\
+      condition:\n\
+        true\n\
+    }", ERROR_INVALID_MODIFIER);
+
+  assert_error(
+    "rule test {\n\
+      strings:\n\
+        $a = \"ab\" base64 fullword\n\
+      condition:\n\
+        true\n\
+    }", ERROR_INVALID_MODIFIER);
+
+  assert_error(
+    "rule test {\n\
+      strings:\n\
+        $a = \"ab\" base64(\"AXS\")\n\
+      condition:\n\
+        true\n\
+    }", ERROR_INVALID_MODIFIER);
+
+  assert_error(
+    "rule test {\n\
+      strings:\n\
+        $a = \"ab\" base64wide(\"ERS\")\n\
+      condition:\n\
+        true\n\
+    }", ERROR_INVALID_MODIFIER);
+
+  // Specifying different alphabets is an error.
+  assert_error(
+    "rule test {\n\
+      strings:\n\
+        $a = \"ab\" base64 base64wide(\"abcdefghijklmnopqrstuvwxyz0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ,.\")\n\
+      condition:\n\
+        true\n\
+    }", ERROR_INVALID_MODIFIER);
+
+  // Be specific about the offsets in these tests to make sure we are matching
+  // the correct strings. Also be specific about the length because we want to
+  // make sure the match is not the entire base64 string, but just the
+  // substrings which are not dependent upon leading or trailing bytes.
+  assert_true_rule_file(
+      "rule test {\n\
+        strings:\n\
+          $a = \"This program cannot\" base64\n\
+        condition:\n\
+          #a == 6 and\n\
+          @a[1] == 0x53 and\n\
+          !a[1] == 25 and\n\
+          @a[2] == 0x70 and\n\
+          !a[2] == 25 and\n\
+          @a[3] == 0xa2 and\n\
+          !a[3] == 24 and\n\
+          @a[4] == 0xbd and\n\
+          !a[4] == 24 and\n\
+          @a[5] == 0xef and\n\
+          !a[5] == 25 and\n\
+          @a[6] == 0x109 and\n\
+          !a[6] == 25\n\
+      }", "tests/data/base64");
+
+  // This is identical to "base64" alone, but test it to make sure we don't
+  // accidentally include the plaintext in the base64 search.
+  assert_true_rule_file(
+      "rule test {\n\
+        strings:\n\
+          $a = \"This program cannot\" base64 ascii\n\
+        condition:\n\
+          #a == 6 and\n\
+          @a[1] == 0x53 and\n\
+          !a[1] == 25 and\n\
+          @a[2] == 0x70 and\n\
+          !a[2] == 25 and\n\
+          @a[3] == 0xa2 and\n\
+          !a[3] == 24 and\n\
+          @a[4] == 0xbd and\n\
+          !a[4] == 24 and\n\
+          @a[5] == 0xef and\n\
+          !a[5] == 25 and\n\
+          @a[6] == 0x109 and\n\
+          !a[6] == 25\n\
+      }", "tests/data/base64");
+
+  // Make sure the wide modifier is applied BEFORE the base64 and we do NOT
+  // include the wide plaintext string.
+  assert_true_rule_file(
+      "rule test {\n\
+        strings:\n\
+          $a = \"This program cannot\" base64 wide\n\
+        condition:\n\
+          #a == 6 and\n\
+          @a[1] == 0x1b5 and\n\
+          !a[1] == 50 and\n\
+          @a[2] == 0x1ea and\n\
+          !a[2] == 50 and\n\
+          @a[3] == 0x248 and\n\
+          !a[3] == 50 and\n\
+          @a[4] == 0x27b and\n\
+          !a[4] == 50 and\n\
+          @a[5] == 0x2db and\n\
+          !a[5] == 50 and\n\
+          @a[6] == 0x311 and\n\
+          !a[6] == 50\n\
+      }", "tests/data/base64");
+
+  // Make sure that both wide and ascii are base64 encoded. We can skip the
+  // verbose length and offset checks, since the previous tests cover that.
+  assert_true_rule_file(
+      "rule test {\n\
+        strings:\n\
+          $a = \"This program cannot\" base64 wide ascii\n\
+        condition:\n\
+          #a == 12\n\
+      }", "tests/data/base64");
+
+  // Make sure that the two strings are generated when one ascii byte is
+  // base64 encoded. When stripped, third base64 encoded is null.
+  assert_true_rule_file(
+      "rule test {\n\
+        strings:\n\
+          $a = \"a\" base64\n\
+          $b = \"a\" base64wide\n\
+        condition:\n\
+          @a[58] == 0x6ac and\n\
+          @a[59] == 0x6b9 and\n\
+          @b[15] == 0x6f7 and\n\
+          @b[16] == 0x711\n\
+      }", "tests/data/base64");
+
+  // In the future, assert false if character classes are generated instead
+  // of stripping the leading and trailing characters
+  assert_true_rule_file(
+      "rule test {\n\
+        strings:\n\
+          $a = \"Dhis program cannow\" base64\n\
+        condition:\n\
+          #a == 2 and\n\
+          @a[1] == 0xa2 and\n\
+          @a[2] == 0xbd\n\
+      }", "tests/data/base64");
+
+  // This checks for the ascii string in base64 form then widened.
+  assert_true_rule_file(
+      "rule test {\n\
+        strings:\n\
+          $a = \"This program cannot\" base64wide\n\
+        condition:\n\
+          #a == 3 and\n\
+          @a[1] == 0x379 and\n\
+          !a[1] == 50 and\n\
+          @a[2] == 0x3b6 and\n\
+          !a[2] == 48 and\n\
+          @a[3] == 0x3f1 and\n\
+          !a[3] == 50\n\
+      }", "tests/data/base64");
+
+  // Logically identical to the test above but include it to make sure we don't
+  // accidentally include the plaintext in the future.
+  assert_true_rule_file(
+      "rule test {\n\
+        strings:\n\
+          $a = \"This program cannot\" base64wide ascii\n\
+        condition:\n\
+          #a == 3 and\n\
+          @a[1] == 0x379 and\n\
+          !a[1] == 50 and\n\
+          @a[2] == 0x3b6 and\n\
+          !a[2] == 48 and\n\
+          @a[3] == 0x3f1 and\n\
+          !a[3] == 50\n\
+      }", "tests/data/base64");
+
+  // Make sure the wide string is base64wide encoded.
+  assert_true_rule_file(
+      "rule test {\n\
+        strings:\n\
+          $a = \"This program cannot\" base64wide wide\n\
+        condition:\n\
+          #a == 3 and\n\
+          @a[1] == 0x458 and\n\
+          !a[1] == 100 and\n\
+          @a[2] == 0x4c5 and\n\
+          !a[2] == 100 and\n\
+          @a[3] == 0x530 and\n\
+          !a[3] == 100\n\
+      }", "tests/data/base64");
+
+  // Make sure both ascii and wide strings are base64wide encoded properly.
+  assert_true_rule_file(
+      "rule test {\n\
+        strings:\n\
+          $a = \"This program cannot\" base64wide wide ascii\n\
+        condition:\n\
+          #a == 6 and\n\
+          @a[1] == 0x379 and\n\
+          !a[1] == 50 and\n\
+          @a[2] == 0x3b6 and\n\
+          !a[2] == 48 and\n\
+          @a[3] == 0x3f1 and\n\
+          !a[3] == 50 and\n\
+          @a[4] == 0x458 and\n\
+          !a[4] == 100 and\n\
+          @a[5] == 0x4c5 and\n\
+          !a[5] == 100 and\n\
+          @a[6] == 0x530 and\n\
+          !a[6] == 100\n\
+      }", "tests/data/base64");
+
+  // Make sure base64 and base64wide together work.
+  assert_true_rule_file(
+      "rule test {\n\
+        strings:\n\
+          $a = \"This program cannot\" base64 base64wide\n\
+        condition:\n\
+          #a == 9 and\n\
+          @a[1] == 0x53 and\n\
+          !a[1] == 25 and\n\
+          @a[2] == 0x70 and\n\
+          !a[2] == 25 and\n\
+          @a[3] == 0xa2 and\n\
+          !a[3] == 24 and\n\
+          @a[4] == 0xbd and\n\
+          !a[4] == 24 and\n\
+          @a[5] == 0xef and\n\
+          !a[5] == 25 and\n\
+          @a[6] == 0x109 and\n\
+          !a[6] == 25 and\n\
+          @a[7] == 0x379 and\n\
+          !a[7] == 50 and\n\
+          @a[8] == 0x3b6 and\n\
+          !a[8] == 48 and\n\
+          @a[9] == 0x3f1 and\n\
+          !a[9] == 50\n\
+      }", "tests/data/base64");
+
+  // Identical to the test above but useful to make sure we don't accidentally
+  // include the ascii plaintext in the future.
+  assert_true_rule_file(
+      "rule test {\n\
+        strings:\n\
+          $a = \"This program cannot\" base64 base64wide ascii\n\
+        condition:\n\
+          #a == 9\n\
+      }", "tests/data/base64");
+
+  // Making sure we don't accidentally include the wide plaintext in the future.
+  assert_true_rule_file(
+      "rule test {\n\
+        strings:\n\
+          $a = \"This program cannot\" base64 base64wide wide\n\
+        condition:\n\
+          #a == 9\n\
+      }", "tests/data/base64");
+
+  assert_true_rule_file(
+      "rule test {\n\
+        strings:\n\
+          $a = \"This program cannot\" base64(\"!@#$\%^&*(){}[].,|ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstu\")\n\
+        condition:\n\
+          #a == 3 and\n\
+          @a[1] == 0x619 and\n\
+          !a[1] == 25 and\n\
+          @a[2] == 0x638 and\n\
+          !a[2] == 24 and\n\
+          @a[3] == 0x656 and\n\
+          !a[3] == 25\n\
+      }", "tests/data/base64");
 }
 
 
@@ -626,26 +1060,6 @@ static void test_hex_strings()
         condition: $a }",
       "1234567890");
 
-  assert_true_rule(
-      "rule test { \
-        strings: $a = { 31 32 [-] // Inline comment\n\r \
-          38 39 } \
-        condition: $a }",
-      "1234567890");
-
-  assert_true_rule(
-      "rule test { \
-        strings: $a = { 31 32 /* Inline comment */ [-] 38 39 } \
-        condition: $a }",
-      "1234567890");
-
-  assert_true_rule(
-      "rule test { \
-        strings: $a = { 31 32 /* Inline multi-line\n\r \
-                                 comment */ [-] 38 39 } \
-        condition: $a }",
-      "1234567890");
-
   assert_true_rule(
       "rule test { \
         strings: $a = {\n 31 32 [-] 38 39 \n\r} \
@@ -658,6 +1072,13 @@ static void test_hex_strings()
         condition: $a }",
       "1234567890");
 
+  assert_true_rule(
+      "rule test { \
+        strings: $a = { 31 32 [-] 33 34 [-] 38 39 } private \
+        condition: $a }",
+      "1234567890");
+
+
   assert_true_rule(
       "rule test { \
         strings: $a = { 31 32 [1] 34 35 [2] 38 39 } \
@@ -804,6 +1225,10 @@ static void test_count()
   assert_true_rule(
       "rule test { strings: $a = \"ssi\" condition: #a == 2 }",
       "mississippi");
+
+  assert_true_rule(
+      "rule test { strings: $a = \"ssi\" private condition: #a == 2 }",
+      "mississippi");
 }
 
 
@@ -815,6 +1240,12 @@ static void test_at()
         condition: $a at 2 and $a at 5 }",
       "mississippi");
 
+  assert_true_rule(
+      "rule test { \
+        strings: $a = \"ssi\" private \
+        condition: $a at 2 and $a at 5 }",
+      "mississippi");
+
   assert_true_rule(
       "rule test { \
         strings: $a = \"mis\" \
@@ -836,6 +1267,12 @@ static void test_in()
         strings: $a = { 6a 2a 58 c3 } \
         condition: $a in (entrypoint .. entrypoint + 1) }",
       PE32_FILE);
+
+  assert_true_rule_blob(
+      "rule test { \
+        strings: $a = { 6a 2a 58 c3 } private \
+        condition: $a in (entrypoint .. entrypoint + 1) }",
+      PE32_FILE);
 }
 
 
@@ -845,6 +1282,10 @@ static void test_offset()
       "rule test { strings: $a = \"ssi\" condition: @a == 2 }",
       "mississippi");
 
+  assert_true_rule(
+      "rule test { strings: $a = \"ssi\" private condition: @a == 2 }",
+      "mississippi");
+
   assert_true_rule(
       "rule test { strings: $a = \"ssi\" condition: @a == @a[1] }",
       "mississippi");
@@ -861,6 +1302,10 @@ static void test_length()
       "rule test { strings: $a = /m.*?ssi/ condition: !a == 5 }",
       "mississippi");
 
+  assert_true_rule(
+      "rule test { strings: $a = /m.*?ssi/ private condition: !a == 5 }",
+      "mississippi");
+
   assert_true_rule(
       "rule test { strings: $a = /m.*?ssi/ condition: !a[1] == 5 }",
       "mississippi");
@@ -907,7 +1352,7 @@ static void test_of()
       "mississippi");
 
   assert_true_rule(
-      "rule test { strings: $a = \"ssi\" $b = \"mis\" $c = \"oops\" "
+      "rule test { strings: $a = \"ssi\" $b = \"mis\" private $c = \"oops\" "
       "condition: 1 of them }",
       "mississippi");
 
@@ -982,6 +1427,227 @@ void test_for()
           for all i in (1..#a) : (@a[i] == 5) \
       }",
       "mississippi");
+
+  assert_true_rule(
+      "rule test { \
+        condition: \
+          for any i in (1, 2, 3) : (i <= 1) \
+      }",
+      NULL);
+
+  assert_true_rule(
+      "rule test { \
+        condition: \
+          for all i in (1, 2, 3) : (i >= 1) \
+      }",
+      NULL);
+
+  assert_false_rule(
+      "rule test { \
+        condition: \
+          for all i in (1, 0) : (i != 1) \
+      }",
+      NULL);
+
+  assert_true_rule(
+      "import \"tests\" \
+      rule test { \
+        condition: \
+          for any item in tests.struct_array : ( \
+            item.i == 1 \
+          ) \
+      }",
+      NULL);
+
+  assert_true_rule(
+      "import \"tests\" \
+      rule test { \
+        condition: \
+          for 0 item in tests.struct_array : ( \
+            item.i == 100 \
+          ) \
+      }",
+      NULL);
+
+  assert_true_rule(
+      "import \"tests\" \
+      rule test { \
+        condition: \
+          for any item in tests.integer_array : ( \
+            item == 2 \
+          ) \
+      }",
+      NULL);
+
+  assert_true_rule(
+      "import \"tests\" \
+      rule test { \
+        condition: \
+          for any item in tests.string_array : ( \
+            item == \"bar\" \
+          ) \
+      }",
+      NULL);
+
+  assert_true_rule(
+      "rule test { \
+        condition: \
+          for all i in (3,5,4) : ( \
+            i >= 3 and i <= 5 \
+          ) \
+      }",
+      NULL);
+
+  assert_true_rule(
+      "rule test { \
+        condition: \
+          for all i in (3..5) : ( \
+            i >= 3 and i <= 5 \
+          ) \
+      }",
+      NULL);
+
+  assert_true_rule(
+      "rule test { \
+        condition: \
+          for 2 i in (5..10) : ( \
+            i == 6 or i == 7 \
+          ) \
+      }",
+      NULL);
+
+  assert_false_rule(
+      "import \"tests\" \
+      rule test { \
+        condition: \
+          for any k,v in tests.empty_struct_dict : ( \
+            true \
+          ) \
+      }",
+    NULL);
+
+  assert_false_rule(
+      "import \"tests\" \
+      rule test { \
+        condition: \
+          for all i in (1..tests.undefined.i) : ( \
+            true \
+          ) \
+      }",
+    NULL);
+
+
+  assert_false_rule(
+      "import \"tests\" \
+      rule test { \
+        condition: \
+          for all i in (tests.undefined.i..10) : ( \
+            true \
+          ) \
+      }",
+    NULL);
+
+  assert_false_rule(
+      "import \"tests\" \
+      rule test { \
+        condition: \
+          for all i in (1..tests.undefined.i) : ( \
+            false \
+          ) \
+      }",
+    NULL);
+
+  // Test case for issue #1180.
+  assert_true_rule(
+      "import \"tests\" \
+      rule test { \
+        condition: \
+          not tests.string_array[4] contains \"foo\" \
+      }",
+    NULL);
+
+  // Test case for issue #1180.
+  assert_true_rule(
+      "import \"tests\" \
+      rule test { \
+        condition: \
+          not tests.string_dict[\"undefined\"] matches /foo/ \
+      }",
+    NULL);
+
+  assert_false_rule(
+      "import \"tests\" \
+      rule test { \
+        condition: \
+          not tests.undefined.i \
+      }",
+    NULL);
+
+  assert_true_rule(
+      "import \"tests\" \
+      rule test { \
+        condition: \
+          for any k,v in tests.struct_dict : ( \
+            k == \"foo\" and v.s == \"foo\" and v.i == 1 \
+          ) \
+      }",
+      NULL);
+
+  assert_error(
+      "import \"tests\" \
+      rule test { \
+        condition: \
+          for any k,v in tests.integer_array : ( false ) \
+      }",
+      ERROR_SYNTAX_ERROR);
+
+  assert_error(
+      "import \"tests\" \
+      rule test { \
+        condition: \
+          for any a,b,c in tests.struct_dict : ( false ) \
+      }",
+      ERROR_SYNTAX_ERROR);
+
+  assert_error(
+      "import \"tests\" \
+      rule test { \
+        condition: \
+          for any i in tests.struct_dict : ( false ) \
+      }",
+      ERROR_SYNTAX_ERROR);
+
+  assert_error(
+      "import \"tests\" \
+      rule test { \
+        condition: \
+          for any i in tests.integer_array : ( undefined_ident ) \
+      }",
+      ERROR_UNDEFINED_IDENTIFIER);
+
+  assert_error(
+      "import \"tests\" \
+      rule test { \
+        condition: \
+          for any i in tests.integer_array : ( i == \"foo\" ) \
+      }",
+      ERROR_WRONG_TYPE);
+
+
+  assert_false_rule(
+      "rule test { \
+        condition: \
+          for any i in (0,1): ( \
+            for any j in (0,1): ( \
+              for any k in (0,1): ( \
+                for any l in (0,1): (\
+                  false \
+                ) \
+              ) \
+            ) \
+        ) \
+      }",
+      NULL);
 }
 
 
@@ -1142,6 +1808,7 @@ void test_re()
   assert_true_regexp("(a+|b)*", "ab", "ab");
   assert_true_regexp("a|b|c|d|e", "e", "e");
   assert_true_regexp("(a|b|c|d|e)f", "ef", "ef");
+  assert_true_regexp("a|b", "a", "a");
   assert_true_regexp(".b{2}", "abb", "abb");
   assert_true_regexp(".b{2,3}", "abbb", "abbb");
   assert_true_regexp(".b{2,3}?", "abbb", "abb");
@@ -1200,6 +1867,10 @@ void test_re()
   assert_true_regexp("ab{1,3}?", "abbbbb", "ab");
   assert_true_regexp("ab{2,2}?", "abbbbb", "abb");
   assert_true_regexp("ab{2,3}?", "abbbbb", "abb");
+  assert_true_regexp("(a{2,3}b){2,3}", "aabaaabaab", "aabaaabaab");
+  assert_true_regexp("(a{2,3}?b){2,3}?", "aabaaabaab", "aabaaab");
+  assert_false_regexp("(a{4,5}b){4,5}", "aaaabaaaabaaaaab");
+  assert_true_regexp("(a{4,5}b){4,5}", "aaaabaaaabaaaaabaaaaab", "aaaabaaaabaaaaabaaaaab");
   assert_true_regexp(".(abc){0,1}", "xabcabcabcabc", "xabc");
   assert_true_regexp(".(abc){0,2}", "xabcabcabcabc", "xabcabc");
   assert_true_regexp("x{1,2}abcd", "xxxxabcd", "xxabcd");
@@ -1249,12 +1920,16 @@ void test_re()
   assert_true_regexp("a\\sb", "a\nb", "a\nb");
   assert_true_regexp("a\\sb", "a\vb", "a\vb");
   assert_true_regexp("a\\sb", "a\fb", "a\fb");
+  assert_true_regexp("a[\\s]*b", "a \t\r\n\v\fb", "a \t\r\n\v\fb");
+  assert_true_regexp("a[^\\S]*b", "a \t\r\n\v\fb", "a \t\r\n\v\fb");
   assert_false_regexp("a\\Sb", "a b");
   assert_false_regexp("a\\Sb", "a\tb");
   assert_false_regexp("a\\Sb", "a\rb");
   assert_false_regexp("a\\Sb", "a\nb");
   assert_false_regexp("a\\Sb", "a\vb");
   assert_false_regexp("a\\Sb", "a\fb");
+  assert_true_regexp("foo([^\\s]*)", "foobar\n", "foobar");
+  assert_true_regexp("foo([^\\s]*)", "foobar\r\n", "foobar");
   assert_true_regexp("\\n\\r\\t\\f\\a", "\n\r\t\f\a", "\n\r\t\f\a");
   assert_true_regexp("[\\n][\\r][\\t][\\f][\\a]", "\n\r\t\f\a", "\n\r\t\f\a");
   assert_true_regexp("\\x01\\x02\\x03", "\x01\x02\x03", "\x01\x02\x03");
@@ -1342,7 +2017,7 @@ void test_re()
   // Test case for issue #324
   assert_true_regexp("whatever|   x.   x", "   xy   x", "   xy   x");
 
-  // test case for issue #503, \x without two following hex-digits
+  // Test case for issue #503, \x without two following hex-digits
   assert_regexp_syntax_error("\\x0");
   assert_regexp_syntax_error("\\x");
 
@@ -1354,6 +2029,13 @@ void test_re()
   // Test case for issue #682
   assert_true_regexp("(a|\\b)[a]{1,}", "aaaa", "aaaa");
 
+  // Test cases for issue #1018
+  assert_true_regexp("(ba{4}){4,10}", "baaaabaaaabaaaabaaaabaaaa", "baaaabaaaabaaaabaaaabaaaa");
+  assert_true_regexp("(ba{2}a{2}){5,10}", "baaaabaaaabaaaabaaaabaaaa", "baaaabaaaabaaaabaaaabaaaa");
+  assert_true_regexp("(ba{3}){4,10}", "baaabaaabaaabaaabaaa", "baaabaaabaaabaaabaaa");
+  assert_true_regexp("(ba{4}){5,10}", "baaaabaaaabaaaabaaaabaaaa", "baaaabaaaabaaaabaaaabaaaa");
+  assert_false_regexp("(ba{4}){4,10}", "baaaabaaaabaaaa");
+
   // Test for integer overflow in repeat interval
   assert_regexp_syntax_error("a{2977952116}");
 
@@ -1365,6 +2047,11 @@ void test_re()
       "rule test { strings: $a = /[a\\/ condition: $a }",
       ERROR_SYNTAX_ERROR);
 
+  // Test case for issue #996
+  assert_error(
+      "rule test {strings:$=/.{,}? /",
+      ERROR_SYNTAX_ERROR);
+
   assert_true_rule_blob(
       "rule test { \
         strings: $a = /MZ.{300,}t/ \
@@ -1408,6 +2095,16 @@ void test_re()
   assert_true_rule(
        "rule test { strings: $a = /abc[^F]/ condition: $a }",
        "abcd");
+
+  // Test case for issue #1006
+  assert_false_rule_blob(
+       "rule test { strings: $a = \" cmd.exe \" nocase wide condition: $a }",
+       ISSUE_1006);
+
+  // Test case for issue #1117
+  assert_true_rule_blob(
+       "rule test { strings: $a =/abc([^\"\\\\])*\"/ nocase condition: $a }",
+       "abc\xE0\x22");
 }
 
 
@@ -1466,6 +2163,45 @@ static void test_comments()
              true\n\
       }",
       NULL);
+
+  assert_true_rule(
+      "rule test { \
+        strings: $a = { 31 32 [-] // Inline comment\n\r \
+          38 39 } \
+        condition: !a == 9 }",
+      "1234567890");
+
+  assert_true_rule(
+      "rule test { \
+        strings: $a = { 31 32 /* Inline comment */ [-] 38 39 } \
+        condition: !a == 9 }",
+      "1234567890");
+
+  assert_true_rule(
+      "rule test { \
+        strings: $a = { 31 32 /* Inline comment */ [-] 38 39 } \
+                 $b = { 31 32 /* Inline comment */ [-] 35 36 } \
+        condition: (!a == 9) and (!b == 6) }",
+      "1234567890");
+
+  assert_true_rule(
+      "rule test { \
+        strings: $a = { 31 32 /* Inline comment with *asterisks* */ [-] 38 39 } \
+        condition: !a == 9}",
+      "1234567890");
+
+  assert_true_rule(
+      "rule test { \
+        strings: $a = { 31 32 /* Inline multi-line\n\r \
+                                 comment */ [-] 38 39 } \
+        condition: !a == 9 }",
+      "1234567890");
+
+  assert_true_rule(
+      "rule test { \
+        strings: $a = { /*Some*/ 31 /*interleaved*/ [-] /*comments*/ 38 39 } \
+        condition: !a == 9 }",
+      "1234567890");
 }
 
 static void test_matches_operator()
@@ -1762,6 +2498,10 @@ static void test_hash_module()
             and \
           hash.sha256(1, filesize) == \
             \"aaaaf2863e043b9df604158ad5c16ff1adaf3fd7e9fcea5dcb322b6762b3b59a\" \
+            and \
+          hash.crc32(0, filesize) == 0x8587d865 \
+            and \
+          hash.crc32(\"TEST STRING\") == 0x51f9be31 \
       }",
       blob);
 
@@ -1819,12 +2559,28 @@ void test_integer_functions()
 void test_include_files()
 {
   assert_true_rule(
-    "include \"tests/data/baz.yar\" rule t { condition: baz }",
-    NULL);
+      "include \"tests/data/baz.yar\" rule t { condition: baz }",
+      NULL);
 
   assert_true_rule(
-    "include \"tests/data/foo.yar\" rule t { condition: foo }",
-    NULL);
+      "include \"tests/data/foo.yar\" rule t { condition: foo }",
+      NULL);
+}
+
+
+void test_tags()
+{
+  assert_true_rule(
+      "rule test : tag1 { condition: true}",
+      NULL);
+
+  assert_true_rule(
+      "rule test : tag1 tag2 { condition: true}",
+      NULL);
+
+  assert_error(
+      "rule test : tag1 tag1 { condition: true}",
+      ERROR_DUPLICATED_TAG_IDENTIFIER);
 }
 
 
@@ -1832,10 +2588,14 @@ void test_process_scan()
 {
   int pid = fork();
   int status = 0;
-  int matches = 0;
   YR_RULES* rules;
   int rc1, rc2;
 
+  struct COUNTERS counters;
+
+  counters.rules_not_matching = 0;
+  counters.rules_matching = 0;
+
   if (pid == 0)
   {
     /* The string should appear somewhere in the shell's process space. */
@@ -1847,14 +2607,22 @@ void test_process_scan()
   /* Give child process time to initialize */
   sleep(1);
 
-  assert( compile_rule("\
+  status = compile_rule("\
     rule test {\
       strings:\
         $a = { 48 65 6c 6c 6f 2c 20 77 6f 72 6c 64 21 }\
       condition:\
         all of them\
-    }", &rules) == ERROR_SUCCESS);
-  rc1 = yr_rules_scan_proc(rules, pid, 0, count_matches, &matches, 0);
+    }", &rules);
+
+  if (status != ERROR_SUCCESS)
+  {
+    perror("compile_rule");
+    exit(EXIT_FAILURE);
+  }
+
+  rc1 = yr_rules_scan_proc(rules, pid, 0, count, &counters, 0);
+  yr_rules_destroy(rules);
   kill(pid, SIGALRM);
 
   rc2 = waitpid(pid, &status, 0);
@@ -1871,7 +2639,7 @@ void test_process_scan()
 
   switch (rc1) {
   case ERROR_SUCCESS:
-    if (matches == 0)
+    if (counters.rules_matching == 0)
     {
       fputs("Found no matches\n", stderr);
       exit(EXIT_FAILURE);
@@ -1899,11 +2667,6 @@ void test_performance_warnings()
         strings: $a = { 01 ?? } \
         condition: $a }")
 
-  assert_warning(
-      "rule test { \
-        strings: $a = { 01 00 } \
-        condition: $a }")
-
   assert_warning(
       "rule test { \
         strings: $a = { 01 ?? ?? } \
@@ -1914,6 +2677,31 @@ void test_performance_warnings()
         strings: $a = { 01 ?? ?? 02 } \
         condition: $a }")
 
+  assert_warning(
+      "rule test { \
+        strings: $a = { 01 ?? ?2 03 } \
+        condition: $a }")
+
+  assert_warning(
+      "rule test { \
+        strings: $a = { 01 ?? 02 1? } \
+        condition: $a }")
+
+  assert_warning(
+      "rule test { \
+        strings: $a = { 1? 2? 3? } \
+        condition: $a }")
+
+  assert_warning(
+      "rule test { \
+        strings: $a = { 1? 2? 3? 04 } \
+        condition: $a }")
+
+  assert_warning(
+      "rule test { \
+        strings: $a = { 1? ?? 03 } \
+        condition: $a }")
+
   assert_warning(
       "rule test { \
         strings: $a = { 00 01 } \
@@ -1934,24 +2722,29 @@ void test_performance_warnings()
         strings: $a = { 00 00 00 } \
         condition: $a }")
 
+  assert_no_warnings(
+      "rule test { \
+        strings: $a = { 00 00 01 } \
+        condition: $a }")
+
   assert_warning(
       "rule test { \
         strings: $a = { 00 00 00 00 } \
         condition: $a }")
 
-  assert_warning(
+  assert_no_warnings(
       "rule test { \
         strings: $a = { 00 00 00 01 } \
         condition: $a }")
 
   assert_warning(
       "rule test { \
-        strings: $a = { 00 00 01 02 } \
+        strings: $a = { FF FF FF FF } \
         condition: $a }")
 
-  assert_warning(
+  assert_no_warnings(
       "rule test { \
-        strings: $a = { FF FF FF FF } \
+        strings: $a = { 00 00 01 02 } \
         condition: $a }")
 
   assert_no_warnings(
@@ -1978,6 +2771,28 @@ void test_performance_warnings()
        "rule test { \
         strings: $a = { 01 02 } \
         condition: $a }")
+
+  assert_no_warnings(
+       "rule test { \
+        strings: $a = \"foo\" wide \
+        condition: $a }")
+
+  assert_no_warnings(
+       "rule test { \
+        strings: $a = \"MZ\" \
+        condition: $a }")
+
+  assert_warning(
+      "rule test { \
+        strings: $a = \"                    \" xor(0x20) \
+        condition: $a }")
+
+  // This will eventually xor with 0x41 and should cause a warning.
+  assert_warning(
+      "rule test { \
+        strings: $a = \"AAAAAAAAAAAAAAAAAAAA\" xor \
+        condition: $a }")
+
 }
 
 
@@ -2021,6 +2836,7 @@ int main(int argc, char** argv)
   // test_string_io();
   test_entrypoint();
   test_global_rules();
+  test_tags();
 
   #if !defined(USE_WINDOWS_PROC) && !defined(USE_NO_PROC)
   test_process_scan();
@@ -2034,6 +2850,5 @@ int main(int argc, char** argv)
   test_performance_warnings();
 
   yr_finalize();
-
   return 0;
 }
